/* lcs2mix.c *****************************************\
*                                                     *
*  Two-step Godunov-type scheme                       *
*  for 1D unsteady flow of gas mixture                *
*                                                     *
*  November 25, 2000								  *
*  author: Andrei A. Chernousov                       *
*          E-mail: <andrei99@iname.com>               *
*          http://www.geocities.com/andrei_chernousov *
******************************************************/

/* Two-step in time scheme smears fronts of shocks
   However it may be used instead of three-step scheme
   for economy. Otherwise use three-step schemes!
*/

/*
Features:
	Model thermodynamics: hk(T)= h0k + cpk * T; (cpk = const)
	Demonstrates convection with constant velocity with abrupt
	  changes in temperature and composition
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*--- for SIGINT handler ---*/
   /**/
#include <signal.h>
   /* continuation flag (to be changed by SIGINT handler) */
int continFlag = 1;
   /* SIGINT handler itself */
#pragma argsused
void termHandler( int sigType )
{
   if( continFlag-- ); /* 1 -> 0 */
   puts( "User break, terminating..." );
}

  /* alias names for fluxes */
#define F1 _U1
#define F2 _U2
#define F3 _U3

#define N_SPEC 4   /* number of species in the mixture */
int k;             /* index for species */

float h0k[N_SPEC], /* enthalpies of formation */
	  cpk[N_SPEC], /* constant! values of heat capacities at p=const */
	  Rk[N_SPEC],  /* gas constants */
	  /**/
	  H, E,
	  /**/
	  P,  R,  U,  T,  C,  K,  Yk [N_SPEC], K_1,
	 _P, _R, _U, _T, _C, _K, _Yk [N_SPEC], _r ,
	  P_, R_, U_, T_, C_, K_, Yk_[N_SPEC],  r_,
	  PRC,
	  /* properties of the mixture */
	  cp_mix,
	  R_mix,
	  h0_mix,
	  cv_mix;

float
   /* conservative variables */
	 /* in cells */
   *U1[N_SPEC],  *U2,  *U3,
   *U1p[N_SPEC], *U2p, *U3p,
	 /* at cells' interfaces */
   *_U1[N_SPEC], *_U2, *_U3,
   *U1_[N_SPEC], *U2_, *U3_,

   /* primitive variables */
   *y[N_SPEC], /* mass fractions */
   *r, /* density */
   *u, /* velocity */
   *t, /* temperature */
   *e, /* total energy - R*h0 */
   *g, /* ratio of specific heats */
   *c, /* speed of sound */

   /* x spacing and time step */
   deltaX, deltaT, deltaT_X,

   /* finite differences of conservative variables */
   m1, m2, m3,
   w1, w2, w3,

   /* finite differences of characteristic variables */
   M1, M2, M3,
   W1, W2, W3,

   /* modified finite differences of characteristic variables */
   _M1, _M2, _M3,
   _W1, _W2, _W3,

   /* transformation matrix [S-1], etc... */
   S11, S12, S13,
   S21, S22, S23,
   S31, S32, S33,

	/* inverted transformation matrix [S-1] */
   _S11, _S12, _S13,
   _S21, _S22, _S23,
   _S31, _S32, _S33,

   /* for LCS procedure */
   J_p, J_m, J_o,
   al_p, al_m, al_o,
   _jp, _jm, _jo,   jp_, jm_, jo_,
   C_p_,C_m_,C_o_,  C_p, C_m, C_o,

   /* misc. stuff */
   k1, k2, k3,
   AAA, BBB, CCC, DDD;

unsigned LEN, LENN, _i, i, i_;
unsigned long step, numstep;
char Stage = 1, New;
FILE *pF;

/*--- For monotone piecewise-parabolic distribution ---*/
	/* constant coefficients, recommended values 4, 1/3, 1/6 */
#define BB 4.000000000000000
#define C1 0.333333333333333
#define C2 0.166666666666667
	/* limiting function - the best: only necessary multiplication *BB  */
float ay;
/*inline*/ float minmod( float x, float y ) {  /* Chakravarthy, Osher */
	if( x*y<=0 ) return 0;
	ay = ( (y*=BB) > 0 ) ? y: -y;
	if( x > 0 ) return (x<ay) ? x : ay;
	return ( (-x<ay) ? x: -ay );
} /* end minmod() */


/************
*  Array1D  *
************/
float *Array1D( unsigned size )
{
float *x;

   if( ( x = (float *)malloc( size * sizeof(float) ) ) == NULL ) {
	  puts( "Cannot allocate memory" );
	  exit( -1 );
   }
   return x;

} /* end Array1D( ) */


/*************
* Initialize *   Initializations
*************/
void Initialize( void )
{
float  /**/
   Y_A, Y_B, Y_C, Y_D,
   Y_Ab, Y_Bb, Y_Cb, Y_Db, Tb;
unsigned LENb;
char str[80];

   /* loading data from file "lcs2mix.ini"  */
   if( ( pF = fopen( "lcs2mix.ini", "r" ) ) == NULL ) {
	  puts( "Cannot open file \"lcs2mix.ini\"" );
	  exit( -1 );
   }
   i = 0;
   do{
	  if ( fgets( str, 79, pF ) != NULL ) {
		 switch( i ) {
			/* stepping */
			case  0:  deltaX = atof( str ); break; /* cell spacing */
			case  1:  deltaT = atof( str ); break; /* time step */
			case  2:  numstep= atoi( str ); break; /* total number of steps */
			case  3:  New    = (char)atoi( str ); break; /* (1-New/0-Continue)*/
			/* zone of right gas */
			case  4:  LEN    = atoi( str ); break; /* total length, in cells */
			case  5:  T      = atof( str ); break;
			case  6:  P      = atof( str ); break;
			case  7:  Y_A    = atof( str ); break; /* composition of */
			case  8:  Y_B    = atof( str ); break; /* mixture in right zone */
			case  9:  Y_C    = atof( str ); break;
			case 10:  Y_D    = atof( str ); break;
			/* zone of zone of another gas */
			case 11:  LENb   = atoi( str ); break; /* length of left zone */
			case 12:  Tb     = atof( str ); break;
			case 13:  Y_Ab   = atof( str ); break; /* composition of */
			case 14:  Y_Bb   = atof( str ); break; /* mixture in left zone */
			case 15:  Y_Cb   = atof( str ); break;
			case 16:  Y_Db   = atof( str ); break;
			/* thermodynaic properties */
			case 17:  h0k[0] = atof( str ); break; /* enthalpies of formation */
			case 18:  h0k[1] = atof( str ); break;
			case 19:  h0k[2] = atof( str ); break;
			case 20:  h0k[3] = atof( str ); break;
			case 21:   Rk[0] = atof( str ); break; /* gas constants */
			case 22:   Rk[1] = atof( str ); break;
			case 23:   Rk[2] = atof( str ); break;
			case 24:   Rk[3] = atof( str ); break;
			case 25:  cpk[0] = atof( str ); break; /* mean capacities at p=const */
			case 26:  cpk[1] = atof( str ); break;
			case 27:  cpk[2] = atof( str ); break;
			case 28:  cpk[3] = atof( str ); break;
			default: break;
		 } /* end switch() */
		 i++;
	  } /* end if() */
	  else {
		 puts( "Error while reading file \"lcs2mix.ini\"" );
		 exit( 1 );
	  }
   } while( i < 29 );
   fclose( pF );

   LENN = LEN+1;
   deltaT_X = deltaT / deltaX;

   /*-- dynamic memory allocation */
	 /* conservative variables */
   for( k = 0; k < N_SPEC; k++ ) {
	  U1[k]  = Array1D( LEN+2 );
	  U1p[k] = Array1D( LEN+2 );
	 _U1[k]  = Array1D( LEN+1 );
	  U1_[k] = Array1D( LEN+1 );
   }
   U2  = Array1D( LEN+2 );  U3  = Array1D( LEN+2 );
   U2p = Array1D( LEN+2 );  U3p = Array1D( LEN+2 );
  _U2  = Array1D( LEN+1 ); _U3  = Array1D( LEN+1 );
   U2_ = Array1D( LEN+1 );  U3_ = Array1D( LEN+1 );
	 /* pimitive variables */
   for( k = 0; k < N_SPEC; k++ ) y[k]  = Array1D( LEN+2 );
   r = Array1D( LEN+2 );
   u = Array1D( LEN+2 );
   t = Array1D( LEN+2 );
   e = Array1D( LEN+2 );
   g = Array1D( LEN+2 );
   c = Array1D( LEN+2 );

   /* from the beginning */
   if( New == 1 ) {

	  /* initial conditions everywhere */
	  Yk[0] = Y_A;
	  Yk[1] = Y_B;
	  Yk[2] = Y_C;
	  Yk[3] = Y_D;
	  cp_mix = R_mix = h0_mix = 0.;
	  for( k = 0; k < N_SPEC; k++ ) {
		 cp_mix += cpk[k] * Yk[k];
		 R_mix  +=  Rk[k] * Yk[k];
		 h0_mix += h0k[k] * Yk[k];
	  }
	  cv_mix = cp_mix - R_mix;
	  U  = 50.;
	  R = P / ( R_mix * T );
	  E = R * ( h0_mix + cv_mix * T );
	  /* conservative variables */
	  for( i = 1; i <= LEN; i++ ) {
		 for( k = 0; k < N_SPEC; k++ ) U1[k][i] = R * Yk[k];
		 U2[i] = R * U;
		 U3[i] = E + 0.5 * R * U * U;
	  }

	  /* initial conditions in small zone */
	  Yk[0] = Y_Ab;
	  Yk[1] = Y_Bb;
	  Yk[2] = Y_Cb;
	  Yk[3] = Y_Db;
	  cp_mix = R_mix = h0_mix = 0.;
	  for( k = 0; k < N_SPEC; k++ ) {
		 cp_mix += cpk[k] * Yk[k];
		 R_mix  +=  Rk[k] * Yk[k];
		 h0_mix += h0k[k] * Yk[k];
	  }
	  cv_mix = cp_mix - R_mix;
	  U = 50.;
	  R = P / ( R_mix * Tb );
	  E = R * ( h0_mix + cv_mix * Tb );
	  /* conservative variables */
	  for( i = LENb + 1; i <= LENb + LENb; i++ ) {
		 for( k = 0; k < N_SPEC; k++ )
		 U1[k][i] = R * Yk[k];
		 U2[i] = R * U;
		 U3[i] = E + 0.5 * R * U * U;
	  } /* end for() */

	  /* step */
	  step = 0;
   }
   /* load solution arrays from file */
   else {
	  /* open file */
	  if ( ( pF = fopen( "backup", "rb" ) ) == NULL ) {
		 puts( "Cannot open \"backup\"" );
		 exit( 1 );
	  }
	  /* read arrays in binary mode */
	  for( k = 0; k < N_SPEC; k++ )
		 fread( &U1[k][1], LEN * sizeof( float ), 1, pF );
	  fread( &U2[1], LEN * sizeof( float ), 1, pF );
	  fread( &U3[1], LEN * sizeof( float ), 1, pF );
	  /* step */
	  fread( &step, sizeof( unsigned long ), 1, pF );
	  /* close file */
	  fclose( pF );
   } /* end else */

   /* primitive variables anyway */
   for ( i = 1; i < LENN; i++ ) {
	  for( k = 0, R = 0.; k < N_SPEC; k++ ) R += U1[k][i];
	  U = U2[i];
	  E = U3[i];
	  /* mass fractions and properties of the mixture */
	  for( k = 0, cp_mix = R_mix = h0_mix = 0.; k < N_SPEC; k++ ){
		 h0_mix += h0k[k] * ( y[k][i] = U1[k][i] / R );
		 cp_mix += cpk[k] *   y[k][i];
		 R_mix  +=  Rk[k] *   y[k][i];
	  }
	  cv_mix = cp_mix - R_mix;
		 /* density */
	  r[i] = R;
		 /* velocity */
	  u[i] = U / R;
		 /* energy e = (r*E) - (r*h0) */
	  e[i] = E - h0_mix * R;
		 /* temperature, from e = (r*E) - (r*u*u/2) */
	  t[i] = ( e[i] - 0.5 * U * U / R ) / ( R * cv_mix );
		 /* ratio of specific heats */
	  g[i] = cp_mix / cv_mix;
		 /* speed of sound */
	  c[i] = sqrt( g[i] * R_mix * t[i] );
   } /* end for() */

} /* end Initialize( ) */


/***********************
* BounCondInGhostCells *
***********************/
void BounCondInGhostCells( void )
{
/* reflective BC
  for( k = 0; k < N_SPEC; k++ ) y[k][0] = y[k][1];
  r[0] =   r[1];
  u[0] = - u[1];
  e[0] =   e[1];
  for( k = 0; k < N_SPEC; k++ ) y[k][LEN+1] = y[k][LEN];
  r[LEN+1] =   r[LEN];
  u[LEN+1] = - u[LEN];
  e[LEN+1] =   e[LEN];
*/
  /* simple convective BC */
  for( k = 0; k < N_SPEC; k++ ) y[k][0] = y[k][1];
  r[0] = r[1];
  u[0] = u[1];
  e[0] = e[1];
  for( k = 0; k < N_SPEC; k++ ) y[k][LEN+1] = y[k][LEN];
  r[LEN+1] = r[LEN];
  u[LEN+1] = u[LEN];
  e[LEN+1] = e[LEN];

} /* end BounCondInGhostCells( ) */


/*****************
* Reconstruction *
*****************/
void Reconstruction( void )
{
/*
   // piecewise-constant reconstruction
   for( _i = 0, i = 1, i_ = 2; i < LENN; _i++, i++, i_++ ) {
	  for( k = 0; k < N_SPEC; k++ ) _U1[k][ i] = U1_[k][_i] = y[k][i] * r[i];
	  _U2[ i] = U2_[_i] = r[i] * u[i];
	  _U3[ i] = U3_[_i] = e[i];
   } // end for()
*/

   /* piecewise-parabolic monotone reconstruction */
   for( _i = 0, i = 1, i_ = 2; i < LENN; _i++, i++, i_++ ) {
	  /*-- at cell centroid */
		/* density */
	  R = r[i];
		/* velocity */
	  U = u[i];
		/* speed of sound */
	  C = c[i];
		/* ratio of specific heats */
	  K = g[i];
	  K_1 = K - 1.;
	  /* inverted transformation matrix [S-1] */
	  AAA = 1/(C+C);
	  BBB = AAA / C;
	  CCC = BBB + BBB;
	  DDD = 0.5 * U * U * BBB;
	  _S11 = -(BBB+BBB); _S12 = _S13 = BBB;
	  _S21 = -U*CCC;     _S22 = _S23 = U * BBB;            _S22 += AAA;               _S23 -= AAA;
	  _S31 = -(DDD+DDD); _S32 = _S33 = DDD + 1./(K_1+K_1); _S32 += ( AAA = U * AAA ); _S33 -= AAA;
	  /* transformation matrix [S] */
	  AAA = 0.5 * K_1 * U * U;
	  BBB = C * U;
	  CCC = - K_1 * U;
	  S11 = -C * C + AAA; S12 = CCC;
	  S21 = - BBB + AAA;  S22 = CCC + C;
	  S31 =   BBB + AAA;  S32 = CCC - C;
	  /* differencing the conservative (gasdynamic) variables */
		 /* forward */
	  m1 = r[i_]         - r[i];
	  m2 = r[i_] * u[i_] - r[i] * u[i];
	  m3 = e[i_]         - e[i];
		  /* backward */
	  w1 = r[i]        - r[_i];
	  w2 = r[i] * u[i] - r[_i] * u[_i];
	  w3 = e[i]        - e[_i];
	  /* finite differences of characteristic variables dW = S * dU */
		 /* forward */
	  W1 = S11*w1 + S12*w2 + K_1*w3;
	  W2 = S21*w1 + S22*w2 + K_1*w3;
	  W3 = S31*w1 + S32*w2 + K_1*w3;
		 /* backward */
	  M1 = S11*m1 + S12*m2 + K_1*m3;
	  M2 = S21*m1 + S22*m2 + K_1*m3;
	  M3 = S31*m1 + S32*m2 + K_1*m3;
	  /* modified characteristic differences _W */
	  _M1 = minmod( M1, W1 );
	  _M2 = minmod( M2, W2 );
	  _M3 = minmod( M3, W3 );
	  _W1 = minmod( W1, M1 );
	  _W2 = minmod( W2, M2 );
	  _W3 = minmod( W3, M3 );
	  /* parameter vector at... */
		 /* rightmost interface */
	  k1 = _M1 * C1 + _W1 * C2;
	  k2 = _M2 * C1 + _W2 * C2;
	  k3 = _M3 * C1 + _W3 * C2;
	  _r      = r[i]        + _S11 * k1 + _S12 * k2 + _S13 * k3;
	  _U2[ i] = r[i] * u[i] + _S21 * k1 + _S22 * k2 + _S23 * k3;
	  _U3[ i] = e[i]        + _S31 * k1 + _S32 * k2 + _S33 * k3;
		 /* leftmost interface */
	  k1 = _M1 * C2 + _W1 * C1;
	  k2 = _M2 * C2 + _W2 * C1;
	  k3 = _M3 * C2 + _W3 * C1;
	  r_      = r[i]        - _S11 * k1 - _S12 * k2 - _S13 * k3;
	  U2_[_i] = r[i] * u[i] - _S21 * k1 - _S22 * k2 - _S23 * k3;
	  U3_[_i] = e[i]        - _S31 * k1 - _S32 * k2 - _S33 * k3;

	  /* reconstruct partial densities at interfaces */
	  for( k = 0; k < N_SPEC; k++ ) {
		 M1 = y[k][i_] - y[k][ i];
		 W1 = y[k][i ] - y[k][_i];
		_M1 = minmod( M1, W1 );
		_W1 = minmod( W1, M1 );
		U1_[k][_i] = r_ * ( y[k][i] - _M1 * C2 - _W1 * C1 );
		_U1[k][ i] = _r * ( y[k][i] + _M1 * C1 + _W1 * C2 );
	  }

   } /* end for() */

} /* end Reconstruction( ) */


/***********************
* BounCondOnInterfaces *  BC at cell boundaries at the domain boundary
***********************/
void BounCondOnInterfaces( void )
{
/* reflective BC
   for( k = 0, R = 0.; k < N_SPEC; k++ ) _U1[k][0] = U1_[k][0];
   _U2[0] = - U2_[0];
   _U3[0] =   U3_[0];
   for( k = 0, R = 0.; k < N_SPEC; k++ ) U1_[k][LEN] = _U1[k][LEN];
   U2_[LEN] = - _U2[LEN];
   U3_[LEN] =   _U3[LEN];
*/

   /* simple convective BC */
   for( k = 0, R = 0.; k < N_SPEC; k++ ) _U1[k][0] = U1_[k][0];
   _U2[0] = U2_[0];
   _U3[0] = U3_[0];
   for( k = 0, R = 0.; k < N_SPEC; k++ ) U1_[k][LEN] = _U1[k][LEN];
   U2_[LEN] = _U2[LEN];
   U3_[LEN] = _U3[LEN];

} /* end BounCondOnInterfaces( void ) */


/***********
*  Fluxes  *
***********/
void Fluxes( void )
{
register float RU;

   for ( i = 0; i < LENN; i++ ) {
	  /*== inviscid fluxes */
	  /*-- left state */
		/* density */
	  _R = 0.;
	  for( k = 0; k < N_SPEC; k++ ) _R += _U1[k][i];
		/* composition */
	  for( k = 0; k < N_SPEC; k++ ) _Yk[k] = _U1[k][i] / _R;
		/* properties of the mixture */
	  cp_mix = R_mix = h0_mix = 0.;
	  for( k = 0; k < N_SPEC; k++ ) {
		 cp_mix += cpk[k] * _Yk[k];
		 R_mix  +=  Rk[k] * _Yk[k];
		 h0_mix += h0k[k] * _Yk[k];
	  }
	  cv_mix = cp_mix - R_mix;
		/* velocity */
	  _U = _U2[i] / _R;
		/* temperature */
	  _T = ( _U3[i] / _R - 0.5 * _U * _U ) / cv_mix;
		/* pressure */
	  _P = _R * R_mix * _T;
		/* ratio cp/cv */
	  _K = cp_mix / cv_mix;
		/* speed of sound */
	  _C = sqrt( _K * R_mix * _T );
		/* "invariants" */
	  _jo = _P - _C * _C * _R;
	  _jp = _U + ( RU = _P / ( _R * _C ) );
	  _jm = _U -   RU;
	  /*-- right state */
		/* density */
	  R_ = 0.;
	  for( k = 0; k < N_SPEC; k++ ) R_ += U1_[k][i];
		/* composition */
	  for( k = 0; k < N_SPEC; k++ ) Yk_[k] = U1_[k][i] / R_;
		/* properties of the mixture */
	  cp_mix = R_mix = h0_mix = 0.;
	  for( k = 0; k < N_SPEC; k++ ) {
		 cp_mix += cpk[k] * Yk_[k];
		 R_mix  +=  Rk[k] * Yk_[k];
		 h0_mix += h0k[k] * Yk_[k];
	  }
	  cv_mix = cp_mix - R_mix;
		/* velocity */
	  U_ = U2_[i] / R_;
		/* temperature */
	  T_ = ( U3_[i] / R_ - 0.5 * U_ * U_ ) / cv_mix;
		/* pressure */
	  P_ = R_ * R_mix * T_;
		/* ratio cp/cv */
	  K_ = cp_mix / cv_mix;
		/* speed of sound */
	  C_ = sqrt( K_ * R_mix * T_ );
		/* "invariants" */
	  jo_ = P_ - C_ * C_ * R_;
	  jp_ = U_ + ( RU = P_ / ( R_ * C_ ) );
	  jm_ = U_ -   RU;
	  /*-- linearized procedure */
	  C_p = ( U = 0.5 * ( _U + U_ ) ) + ( C = 0.5 * ( _C + C_ ) );
	  C_m = U - C;
	  C_o = U;
		/* J_p and al_p */
	  if( C_p > 0. ) { al_p = 0.5 * ( _K - 1. ) * ( _jm - _jp ) / _jo; J_p  = _jp; }
	  else           { al_p = 0.5 * ( K_ - 1. ) * ( jm_ - jp_ ) / jo_; J_p  = jp_; }
		/* J_m and al_m */
	  if( C_m > 0. ) { al_m = 0.5 * ( _K - 1. )  * ( _jp - _jm ) / _jo; J_m  = _jm; }
	  else           { al_m = 0.5 * ( K_ - 1. )  * ( jp_ - jm_ ) / jo_; J_m  = jm_; }
		/* J_o and al_o */
	  if( C_o > 0. ) {
		 al_o = 0.5 * _K * ( _jp - _jm );
		 al_o = - al_o * al_o;
		 J_o  = _jo;
		 for( k = 0; k < N_SPEC; k++ ) Yk[k] = _Yk[k];
	  }
	  else {
		 al_o = 0.5 * K_ * ( jp_ - jm_ );
		 al_o = - al_o * al_o;
		 J_o  = jo_;
		 for( k = 0; k < N_SPEC; k++ ) Yk[k] = Yk_[k];
	  }
		/* parameters at the interface */
	  P = ( J_p - J_m ) / ( al_p - al_m );
	  U = J_p - al_p * P;
	  R = ( J_o - P ) / al_o;
		/* properties */
	  cp_mix = R_mix = h0_mix = 0.;
	  for( k = 0; k < N_SPEC; k++ ) {
		 cp_mix += cpk[k] * Yk[k];
		 R_mix  +=  Rk[k] * Yk[k];
		 h0_mix += h0k[k] * Yk[k];
	  }
	  cv_mix = cp_mix - R_mix;
		/* temperature from EOS */
	  T = P / ( R * R_mix );
		/* enthalpy - with enthalpy of formation added */
	  H = h0_mix + cp_mix * T;
	  /*-- fluxes */
	  RU = R * U;
	  for( k = 0; k < N_SPEC; k++ )
	  F1[k][i] = RU * Yk[k];
	  F2[i]    = RU * U + P;
	  F3[i]    = RU * ( H + 0.5 * U * U );
   } /* end for() */

} /* end Fluxes( ) */


/**************
*  Evolution  *   Evilution stage for dependent variables
**************/
void Evolution( void )
{
   if( Stage == 1 ) {
	  for ( _i = 0, i = 1; i < LENN; i++, _i++ ) {
		 /* conservative variables */
		 for( k = 0, R = 0.; k < N_SPEC; k++ )
		 R += ( U1p[k][i] = U1[k][i] + deltaT_X * ( F1[k][_i] - F1[k][i] ) );
		 U  =   U2p[i] =    U2[i]    + deltaT_X * ( F2[_i]    - F2[i]    );
		 E  =   U3p[i] =    U3[i]    + deltaT_X * ( F3[_i]    - F3[i]    );
		 /* primitive (gasdynamic) variables */
		   /* mass fractions and properties of the mixture */
		 for( k = 0, cp_mix = R_mix = h0_mix = 0.; k < N_SPEC; k++ ){
			h0_mix += h0k[k] * ( y[k][i] = U1p[k][i] / R );
			cp_mix += cpk[k] *   y[k][i];
			R_mix  +=  Rk[k] *   y[k][i];
		 }
		 cv_mix = cp_mix - R_mix;
		   /* density */
		 r[i] = R;
		   /* velocity */
		 u[i] = U / R;
		   /* energy e = (r*E) - (r*h0) */
		 e[i] = E - h0_mix * R;
		   /* temperature, from e = (r*E) - (r*u*u/2) */
		 t[i] = ( e[i] - 0.5 * U * U / R ) / ( R * cv_mix );
		   /* ratio of specific heats */
		 g[i] = cp_mix / cv_mix;
		   /* speed of sound */
		 c[i] = sqrt( g[i] * R_mix * t[i] );
	  } /* end for() */
	  Stage = 2;
   }
   else {  /* Stage == 2 */
	  for ( _i = 0, i = 1; i <= LEN; i++, _i++  ) {
		 /* conservative variables */
		 for( k = 0, R = 0.; k < N_SPEC; k++ )
		 R += ( U1[k][i] = 0.5 * ( U1[k][i] + U1p[k][i] + deltaT_X * ( F1[k][_i] - F1[k][i] ) ) );
		 U  =   U2[i] =    0.5 * ( U2[i]    + U2p[i]    + deltaT_X * ( F2[_i]    - F2[i]    ) );
		 E  =   U3[i] =    0.5 * ( U3[i]    + U3p[i]    + deltaT_X * ( F3[_i]    - F3[i]    ) );
		 /* primitive (gasdynamic) variables */
		   /* mass fractions and properties of the mixture */
		 for( k = 0, cp_mix = R_mix = h0_mix = 0.; k < N_SPEC; k++ ){
			h0_mix += h0k[k] * ( y[k][i] = U1[k][i] / R );
			cp_mix += cpk[k] *   y[k][i];
			R_mix  +=  Rk[k] *   y[k][i];
		 }
		 cv_mix = cp_mix - R_mix;
		   /* density */
		 r[i] = R;
		   /* velocity */
		 u[i] = U / R;
		   /* energy e = (r*E) - (r*h0) */
		 e[i] = E - h0_mix * R;
		   /* temperature, from e = (r*E) - (r*u*u/2) */
		 t[i] = ( e[i] - 0.5 * U * U / R ) / ( R * cv_mix );
		   /* ratio of specific heats */
		 g[i] = cp_mix / cv_mix;
		   /* speed of sound */
		 c[i] = sqrt( g[i] * R_mix * t[i] );
	  } /* end for() */
	  Stage = 1;
   } /* end else */
} /* end Evolution() */


/***********
*  Output  *    Write solution to "input.dat" file
***********/
void Output( void )
{
   /* open file */
   if ( ( pF = fopen( "input.dat", "w" ) ) == NULL ) {
	  puts( "Cannot open file\"input.dat\"" );
	  exit( -1 );
   }
	 /* header */
   fprintf( pF, "%d\n%d\n", LEN + 2, N_SPEC + 2 );
	 /* first scaling line */
   fprintf( pF, "0. " );
   for( k = 0; k < N_SPEC; k++ ) fprintf( pF, "-0.1 " );
   fprintf( pF, "0. 0. 0. \n" );
	 /* parameters in cells */
   for ( i = 1; i <= LEN; i++ ) {
		/* coordinate */
	  fprintf( pF, "%1.6f ", (i-0.5)*deltaX );
		/* composition */
	  for( k = 0; k < N_SPEC; k++ )
		 fprintf( pF, "%1.4f ", y[k][i] );
		/* density, temperature, velocity */
	  fprintf( pF, "%1.3f %4.1f %2.3f \n", r[i], t[i], u[i] );
   }
	 /* last scaling line */
   fprintf( pF, "%f ", LEN * deltaX );
   for( k = 0; k < N_SPEC; k++ ) fprintf( pF, "1.1 " );
   fprintf( pF, "1.5 3000. 100. \n " );
	 /* close file */
   fclose( pF );
   
} /* end Output() */


/*************
*  Finalize  *    Save solution to binary "backup" file
*************/
void Finalize( void )
{
   /* open file */
   if ( ( pF = fopen( "backup", "wb" ) ) == NULL ) {
	  puts( "Cannot open file \"backup\"" );
	  exit( -1 );
   }
   /* write arrays in binary mode */
   for( k = 0; k < N_SPEC; k++ )
	  fwrite( &U1[k][1], LEN * sizeof( float ), 1, pF );
   fwrite( &U2[1], LEN * sizeof( float ), 1, pF );
   fwrite( &U3[1], LEN * sizeof( float ), 1, pF );
   fwrite( &step, sizeof( unsigned long ), 1, pF );
	 /* close file */
   fclose( pF );
   
} /* end Finalize() */


/*********
*  MAIN  *
**********/
int main( void )
{
   /*--- a handler for a signal SIGINT ---*/
   signal( SIGINT, termHandler );
   
   /*--- initialization ---*/
   Initialize();
   
   /*---*/
   while ( 1 ) {
   
	  /*--- check exit conditions ---*/
	  if( step >= numstep || !continFlag  ) break;
	  printf( "%lu\n", ++step );
	  
	  /*=== first stage ===*/
	  /*--- BC in ghost cells ---*/
	  BounCondInGhostCells( );
	  /*--- parameters at cell interfaces ---*/
	  Reconstruction( );
	  /*---*/
	  BounCondOnInterfaces( );
	  /*--- fluxes at cell interfaces ---*/
	  Fluxes( );
	  /*--- evolution ---*/
	  Evolution( );
	  
	  /*=== second stage ===*/
	  /*--- BC in ghost cells ---*/
	  BounCondInGhostCells( );
	  /*--- parameters at cell interfaces ---*/
	  Reconstruction( );
	  /*---*/
	  BounCondOnInterfaces( );
	  /*--- fluxes at cell interfaces ---*/
	  Fluxes( );
	  /*--- evolution ---*/
	  Evolution( );
	  
   } /* end while */
   
   /*--- output flowfield ---*/
   Output();
   
   /*--- save solution to binary file ---*/
   Finalize();
   
   return 0;
   
}  /* end main()*/

/* end lcs2mix.c */